"""Define GraphQL schema."""

import sys
from typing import Any, Dict, Optional

from graphenum.entities.graphql.properties import InputObjectName, InterfaceName, ObjectName, UnionName
from graphenum.entities.graphql.types import GraphEnum, GraphInputObject, GraphInterface, GraphObject, GraphUnion

if sys.version_info >= (3, 8):
    from typing import TypedDict  # pylint: disable=no-name-in-module
else:
    from typing_extensions import TypedDict


class IntrospectionSchema(TypedDict):

    """Define a GraphQL Introspection Schema.

    Source:
        https://spec.graphql.org/draft/#sec-Schema-Introspection.Schema-Introspection-Schema
    """

    query: Optional[GraphObject]
    mutation: Optional[GraphObject]
    subscription: Optional[GraphObject]
    directives: Dict
    objects: Dict[ObjectName, GraphObject]
    inputObjects: Dict[InputObjectName, GraphInputObject]
    unions: Dict[UnionName, GraphUnion]
    interfaces: Dict[InterfaceName, GraphInterface]
    enums: Dict[str, GraphEnum]
    others: Dict


class RawSchema(TypedDict):

    """Define a GraphQL Raw Schema."""

    query: Any
    mutation: Any
    subscription: Any
    directives: Any
    objects: Any
    inputObjects: Any
    others: Any
    unions: Any
    interfaces: Any
    scalars: Any
    enums: Any



"""Manage serialization/deserialization of the schema."""

from typing import Any, Dict, List, Optional, Set, cast

from graphenum.entities.graphql import GraphType, Operation, RawSchema, SpecKey
from graphenum.logger import LOG

RESERVED_KEYWORDS = ['__Schema', '__Type', '__Field', '__InputValue', '__EnumValue', '__Directive']


def _init_fields(dct: dict) -> dict:
    """Initialize a field in the schema."""

    for field in dct['fields']:

        field['pii'] = []

        for arg in field['args']:
            arg['sources'] = []

    return dct


# pylint: disable=too-many-branches
def serialize_schema(gql_schema: Dict) -> RawSchema:
    """Serialize a GraphQL schema to a RawSchema."""

    raw_schema = RawSchema({
        'query': None,
        'mutation': None,
        'subscription': None,
        'directives': {},
        'objects': {},
        'inputObjects': {},
        'unions': {},
        'interfaces': {},
        'others': {},
        'scalars': {},
        'enums': {}
    })

    document_types: Dict[Operation, Set[str]] = {
        'query': set(),
        'mutation': set(),
        'subscription': set(),
    }

    for operation, operation_names in document_types.items():
        operation_type = f'{operation}Type'
        if gql_schema.get(operation_type):
            operation_names.add(gql_schema[operation_type].get('name', '').lower())
        else:
            LOG.info(f'No {operation_type} detected')

    for typ in gql_schema['types']:

        if typ['kind'] == 'OBJECT':

            if typ['name'].lower().startswith('__'):  # For the moment we don't want to deal with reserved keywords
                continue

            # We want to save sent arguments in each field of queries, mutations, and subcriptions…
            is_operation = False
            for operation, operation_names in document_types.items():
                if typ['name'].lower() in operation_names:
                    raw_schema[operation] = _init_fields(typ)
                    is_operation = True
                    break

            # … but also normal objects.
            # In addition, we want to save the returned values of objects when querrying them.
            if not is_operation:
                raw_schema['objects'][typ['name']] = _init_fields(typ)
                raw_schema['objects'][typ['name']]['returnedExamples'] = {}
                raw_schema['objects'][typ['name']]['paths'] = {'query': {}, 'mutation': {}}

        # Input objects have no fields but directly arguments.
        # They cannont be used as a return type.
        elif typ['kind'] == 'INPUT_OBJECT':
            for input_field in typ['inputFields']:
                input_field['sources'] = []
            raw_schema['inputObjects'][typ['name']] = typ

        elif typ['kind'] == 'UNION':
            raw_schema['unions'][typ['name']] = typ
            typ['possibleTypes'] = {possible_type['name']: possible_type for possible_type in typ['possibleTypes']}
            raw_schema['unions'][typ['name']]['paths'] = {'query': {}, 'mutation': {}}

        elif typ['kind'] == 'SCALAR':
            raw_schema['scalars'][typ['name']] = {'name': typ['name'], 'description': typ['description'], 'returnedExamples': []}

        elif typ['kind'] == 'INTERFACE':
            raw_schema['interfaces'][typ['name']] = typ
            typ['possibleTypes'] = {possible_type['name']: possible_type for possible_type in typ['possibleTypes']}
            raw_schema['interfaces'][typ['name']]['paths'] = {'query': {}, 'mutation': {}}

        elif typ['kind'] == 'ENUM':
            raw_schema['enums'][typ['name']] = {'name': typ['name'], 'description': typ['description'], 'enumValues': typ['enumValues']}

        # We do not want to save anything for other values so far.
        else:
            raw_schema['others'][typ['name']] = typ

    return raw_schema


def _process_raw_type(graph_type: dict) -> Optional[GraphType]:
    """Recursively flatten types.
    Note: at the beginning of this function, the "graph_type" is still not a valid graph_type.

    TODO: Might need to find a better solution for complex types (accession…)
    """
    if graph_type['kind'] in ['OBJECT', 'SCALAR', 'ENUM', 'INTERFACE', 'INPUT_OBJECT', 'UNION']:
        return GraphType(name=graph_type['name'], nullable=[True], depth=0, kind=graph_type['kind'])

    if graph_type['kind'] == 'LIST':
        of_type = _process_raw_type(graph_type['ofType'])
        of_type = cast(GraphType, of_type)
        of_type['depth'] += 1
        of_type['nullable'] = [True] + of_type['nullable']
        return of_type

    if graph_type['kind'] == 'NON_NULL':
        of_type = _process_raw_type(graph_type['ofType'])
        of_type = cast(GraphType, of_type)
        of_type['nullable'][0] = False
        return of_type

    # If this raises, check if we forgot to support one type
    LOG.warning(f'graphenum [TODO]: Unsupported field : {graph_type}')
    return None


def flatten_types(specification: RawSchema) -> RawSchema:  #pylint: disable=too-many-branches
    """Flatten types everywhere in the spec.
    Note: at the beginning of this function, the "specification" is still not a valid specification.

    # TODO: Factorize this
    """

    LOG.info('Recursively flattening schema types to custom GraphTypes')
    for key in specification:

        key = cast(SpecKey, key)

        # We don't want fields that return graphql special type like fields in RESERVED_KEYWORDS
        # We first get the index of element to remove then we remove it.
        if specification[key]:
            index_to_suppress = (index for index, field in enumerate(specification[key].get('fields', [])) if field['type']['name'] in RESERVED_KEYWORDS)
            for index in index_to_suppress:
                specification[key]['fields'].pop(index)

            if key == 'objects':
                for obj in specification['objects'].values():
                    for field in obj['fields']:
                        field['type'] = _process_raw_type(field['type'])
                        for arg in field['args']:
                            arg['type'] = _process_raw_type(arg['type'])

            elif key == 'inputObjects':
                for obj in specification['inputObjects'].values():
                    for input_field in obj['inputFields']:
                        input_field['type'] = _process_raw_type(input_field['type'])

            elif key == 'unions':
                for obj in specification['unions'].values():
                    for possible_type in obj['possibleTypes'].values():
                        obj['possibleTypes'][possible_type['name']] = _process_raw_type(possible_type)

            elif key == 'interfaces':
                for obj in specification['interfaces'].values():
                    for field in obj['fields']:
                        field['type'] = _process_raw_type(field['type'])
                        for arg in field['args']:
                            arg['type'] = _process_raw_type(arg['type'])
                    for possible_type in obj['possibleTypes'].values():
                        obj['possibleTypes'][possible_type['name']] = _process_raw_type(possible_type)

            else:

                for index, field in enumerate(specification[key].get('fields', [])):
                    field['type'] = _process_raw_type(field['type'])
                    for arg in field['args']:
                        arg['type'] = _process_raw_type(arg['type'])

    return specification


def lists_to_dicts(obj: Any) -> Any:
    """Convert lists of dicts to dicts (by duplicating the "name" to use it as a key) for the keys 'fields' and 'args'."""

    if isinstance(obj, List):
        return list(map(lists_to_dicts, obj))

    if isinstance(obj, Dict):

        for key in ['fields', 'inputFields', 'args', 'enumValues', 'interfaces']:
            if key in obj and isinstance(obj[key], List):
                obj[key] = {el['name']: el for el in obj[key]}

        return {k: lists_to_dicts(v) for k, v in obj.items()}

    return obj



"""Define GraphQL properties."""

import sys
from enum import Enum, unique
from typing import List, NewType, Tuple, Union

from graphenum.utils import MetaEnum

if sys.version_info >= (3, 8):
    from typing import Literal, TypeAlias  # pylint: disable=no-name-in-module
else:
    from typing_extensions import Literal, TypeAlias

Operation: TypeAlias = Literal['query', 'mutation', 'subscription']
SupportedOperations: Tuple[Operation, Operation] = ('query', 'mutation')
GraphValue = Union[int, str, float, bool, List['GraphValue']]  # type: ignore[misc]
SpecKey = Literal['query', 'mutation', 'subscription', 'unions', 'objects', 'interfaces', 'inputObjects']
ObjectType = Literal['unions', 'objects', 'interfaces']

ArgName = NewType('ArgName', str)
FieldName = NewType('FieldName', str)
ObjectName = NewType('ObjectName', str)
InputObjectName = NewType('InputObjectName', str)
UnionName = NewType('UnionName', str)
InterfaceName = NewType('InterfaceName', str)
EnumName = NewType('EnumName', str)
CustomScalarName = NewType('CustomScalarName', str)

GraphObjectType = Union[ObjectName, UnionName, InterfaceName]
GraphPath = Tuple[Tuple[FieldName, GraphObjectType], ...]


@unique
class GraphPrimitive(str, Enum, metaclass=MetaEnum):

    """The default GraphQL Scalar types."""

    ID = 'ID'
    INT = 'Int'
    STRING = 'String'
    BOOLEAN = 'Boolean'
    FLOAT = 'Float'


GraphTypeName = Union[ObjectName, InputObjectName, UnionName, InterfaceName, CustomScalarName, EnumName]


@unique
class GraphKind(str, Enum):

    """GraphQL Kind.

    Source:
        https://spec.graphql.org/draft/#sec-Schema-Introspection.Schema-Introspection-Schema
        -> __TypeKind
    """

    SCALAR = 'SCALAR'
    OBJECT = 'OBJECT'
    INTERFACE = 'INTERFACE'
    UNION = 'UNION'
    ENUM = 'ENUM'
    INPUT_OBJECT = 'INPUT_OBJECT'
    # LIST
    # NON_NULL



"""Export GraphQL entities."""

from graphenum.entities.graphql.properties import *  # noqa
from graphenum.entities.graphql.schema import *  # noqa
from graphenum.entities.graphql.types import *  # noqa



"""Utils.

Theses functions are meant to have no dependencies on the graphintrospect package.
"""

import json
from enum import EnumMeta
from typing import Any, Type, Union


# pylint: disable=no-value-for-parameter
class MetaEnum(EnumMeta):

    """Meta class for Enum."""

    def __contains__(
        cls: Type[Any],
        obj: object,
    ) -> bool:
        """Check if the object is in the enum using `in` keyword."""

        try:
            cls(obj)
        except ValueError:
            return False
        return True


class SetEncoder(json.JSONEncoder):

    """Encode sets as lists."""

    def default(self, o: Any) -> Any:
        """Encode sets as lists."""
        if isinstance(o, set):
            return list(o)
        return json.JSONEncoder.default(self, o)


def dump_schema(path: str, schema: Union[Any, dict]) -> None:
    """Dump the schema to a file."""
    with open(path, 'w', encoding='utf-8') as f:
        f.write(json.dumps(schema, indent=2, cls=SetEncoder))



"""Manage module flow."""

import asyncio
import logging
from typing import Callable, Dict, List, Optional, Tuple

import aiohttp

from graphenum.entities.errors import IntrospectionError
from graphenum.entities.graphql import IntrospectionSchema
from graphenum.entities.internal import Callbacks
from graphenum.logger import LOG, setup_logger  # pylint: disable=unused-import # noqa
from graphenum.paths import generate_objects_paths
from graphenum.serdes import flatten_types, lists_to_dicts, serialize_schema
from graphenum.sources import fetch_schema, load_schema


async def async_introspect(
    url: str,
    logger: Optional[logging.Logger] = None,
    headers: Optional[Dict[str, str]] = None,
    verbose_mode: bool = False,
    schema_path: Optional[str] = None,
    callbacks: Optional[Dict[Callbacks, List[Callable]]] = None,
) -> Tuple[IntrospectionSchema, aiohttp.ClientResponse]:
    """Run the introspection.

    Args:
        url: The URL of the GraphQL endpoint.
        logger: The logger to use.
        headers: The headers to send with the fetch_schema request.
        verbose_mode: Infer logging level.
        schema_path: The path to the provided GraphQL schema.

    Returns:
        The introspected schema.
    """

    callbacks = callbacks or {}

    global LOG  # pylint: disable=global-statement
    LOG = logger or setup_logger(verbose_mode)

    gql_schema, response = await fetch_schema(url, headers)
    if not gql_schema and schema_path:
        gql_schema = load_schema(schema_path)

    assert gql_schema, IntrospectionError(f'Could not retrieve schema from {url}')

    if callbacks.get(Callbacks.PRE_SERIALIZE):
        for callback in callbacks[Callbacks.PRE_SERIALIZE]:
            raw_schema = callback(gql_schema)

    raw_schema = serialize_schema(gql_schema)
    raw_schema = flatten_types(raw_schema)
    if callbacks.get(Callbacks.POST_SERIALIZE):
        for callback in callbacks[Callbacks.POST_SERIALIZE]:
            raw_schema = callback(gql_schema)

    schema = lists_to_dicts(raw_schema)
    if callbacks.get(Callbacks.PRE_SCHEMA_GENERATION):
        for callback in callbacks[Callbacks.PRE_SCHEMA_GENERATION]:
            schema = callback(schema)

    schema = generate_objects_paths(schema)
    if callbacks.get(Callbacks.POST_SCHEMA_GENERATION):
        for callback in callbacks[Callbacks.POST_SCHEMA_GENERATION]:
            schema = callback(schema)

    return schema, response


def introspect(
    url: str,
    logger: Optional[logging.Logger] = None,
    headers: Optional[Dict[str, str]] = None,
    verbose_mode: bool = False,
    schema_path: Optional[str] = None,
    callbacks: Optional[Dict[Callbacks, List[Callable]]] = None,
) -> Tuple[IntrospectionSchema, aiohttp.ClientResponse]:
    """Create a new asyncronous task to run the introspection."""

    return asyncio.run(async_introspect(
        url=url,
        logger=logger,
        headers=headers,
        verbose_mode=verbose_mode,
        schema_path=schema_path,
        callbacks=callbacks,
    ))



"""Manage paths generation."""

import random
from typing import Counter, Dict, List, Optional, Set, Tuple, Union, cast

from graphenum import settings
from graphenum.entities.graphql import FieldName, GraphInterface, GraphObjectType, GraphPath, GraphUnion, InterfaceName, IntrospectionSchema, ObjectName
from graphenum.entities.graphql import ObjectType, Operation, SupportedOperations, UnionName
from graphenum.logger import LOG


def _generate_reverse_objects_dict(schema: IntrospectionSchema) -> Dict[GraphObjectType, List[Tuple[FieldName, GraphObjectType]]]:
    """Generate a Dictionary that associate the name of an object to all the objects that include this object in their fields.

    <object_name> str: List[Tuple[str, str]] [(<field_name_A>, <parent_object_name_A>), (<field_name_B>, <parent_object_name_B>)]
    """

    reverse_dict: Dict[GraphObjectType, List[Tuple[FieldName, GraphObjectType]]] = {}

    for parent_name, obj in schema['objects'].items():

        for field_name, field in obj['fields'].items():
            child_name = field['type']['name']

            if field['type']['kind'] in ['OBJECT', 'UNION', 'INTERFACE']:
                child_name = cast(GraphObjectType, child_name)
                reverse_dict[child_name] = reverse_dict.get(child_name, []) + [(field_name, parent_name)]

            # For unions and interfaces we add a step to select all possible types
            if field['type']['kind'] == 'UNION':
                child_name = cast(UnionName, child_name)
                graph_union: GraphUnion = schema['unions'][child_name]
                for possible_type in graph_union['possibleTypes']:
                    reverse_dict[possible_type] = reverse_dict.get(possible_type, []) + [(FieldName(f'... on {possible_type}'), child_name)]

                # Limit maximum of possible union types to avoid too much recursion.
                # This is mandatory on big applications.
                if settings.MAX_POSSIBLE_TYPES_PER_UNIONS > 0:
                    reverse_dict = {k: reverse_dict[k] for k in list(reverse_dict.keys())[:settings.MAX_POSSIBLE_TYPES_PER_UNIONS]}

            if field['type']['kind'] == 'INTERFACE':
                child_name = cast(InterfaceName, child_name)
                graph_interface: GraphInterface = schema['interfaces'][child_name]
                for possible_type in graph_interface['possibleTypes']:
                    reverse_dict[possible_type] = reverse_dict.get(possible_type, []) + [(FieldName(f'... on {possible_type}'), child_name)]

    return reverse_dict


def _generate_reverse_doc_type_dict(
    schema: IntrospectionSchema,
    operation: Operation,
) -> Dict[GraphObjectType, List[FieldName]]:
    """Generate a dictionary of the queries/mutations that return an object.

    <object_name> str: List[str] [<query_A>, <query_B>, ...]
    """

    reverse_dict: Dict[GraphObjectType, List[FieldName]] = {}

    operation_field = schema[operation]
    if operation_field:
        for field_name, field in operation_field['fields'].items():

            if field['type']['kind'] in ['OBJECT', 'UNION', 'INTERFACE']:
                type_name = cast(GraphObjectType, field['type']['name'])
                reverse_dict[type_name] = reverse_dict.get(type_name, []) + [field_name]

    if not reverse_dict:
        LOG.error(f'It seems there is no paths coming from a "{operation}"')

    return reverse_dict


def _get_leading_paths_object(
    root_parents: List[FieldName],
    object_name: GraphObjectType,
) -> Set[GraphPath]:
    """Get List of the paths that lead to an object."""

    return {((root_parent, object_name), ) for root_parent in root_parents}


def _get_parent_object_type(
    schema: IntrospectionSchema,
    object_parent: Tuple[FieldName, GraphObjectType],
) -> Optional[ObjectType]:
    """Determinate parent object type."""

    if object_parent[1] in schema['objects'].keys():
        return 'objects'
    if object_parent[1] in schema['unions'].keys():
        return 'unions'
    if object_parent[1] in schema['interfaces'].keys():
        return 'interfaces'

    LOG.warning(f'Escape[TODO]: We do not support this type for creating the path: {schema["others"][object_parent[1]]["kind"]}')
    return None


def _generate_parents_object_paths(
    schema: IntrospectionSchema,
    operation: Operation,
    object_parent: Tuple[FieldName, GraphObjectType],
    object_name: GraphObjectType,
    reverse_objects_dict: Dict[GraphObjectType, List[Tuple[FieldName, GraphObjectType]]],
    reverse_doc_type_dict: Dict[GraphObjectType, List[FieldName]],
    already_seen: Counter[GraphObjectType],
) -> Set[GraphPath]:
    """Generate parents paths for an object."""

    parent_node = (object_parent[0], object_name)
    parent_object_type: Optional[ObjectType] = _get_parent_object_type(schema, object_parent)
    if not parent_object_type:
        return set()

    # Get the paths of the parent object, compute them if they don't exist yet
    parent_paths: Set[GraphPath] = schema[parent_object_type][object_parent[1]]['paths'][operation]  # type: ignore[index]
    if not parent_paths:
        _generate_nested_object_paths(
            schema,
            operation,
            object_parent[1],
            parent_object_type,
            reverse_objects_dict,
            reverse_doc_type_dict,
            already_seen.copy(),
        )
        parent_paths = schema[parent_object_type][object_parent[1]]['paths'][operation]  # type: ignore[index]

    paths: Set[GraphPath] = set()
    for parent_path in parent_paths:
        if object_name in [parent_name for _, parent_name in parent_path]:
            continue
        paths.add(parent_path + (parent_node, ))

    return paths


def _generate_nested_object_paths(
    schema: IntrospectionSchema,
    operation: Operation,
    object_name: Union[ObjectName, UnionName, InterfaceName],
    object_type: ObjectType,
    reverse_objects_dict: Dict[Union[ObjectName, UnionName, InterfaceName], List[Tuple[FieldName, Union[ObjectName, UnionName, InterfaceName]]]],
    reverse_doc_type_dict: Dict[Union[ObjectName, UnionName, InterfaceName], List[FieldName]],
    already_seen: Counter[Union[ObjectName, UnionName, InterfaceName]] = None,
) -> IntrospectionSchema:
    """Recursively save the paths that lead to one object."""

    # If the parent object has already been treated, we skip it to avoid endless loops
    already_seen = already_seen or Counter()
    if already_seen[object_name] > 0:
        return schema
    already_seen[object_name] += 1

    root_parents = reverse_doc_type_dict.get(object_name, [])  # List of the queries/mutations that return this object
    object_parents = reverse_objects_dict.get(object_name, [])  # List of the objects that contain this object in their fields
    paths: Set[GraphPath] = _get_leading_paths_object(root_parents, object_name)

    # Then look for the "parent" objects that contain this object in their fields
    for object_parent in object_parents:

        parent_paths: Set[GraphPath] = _generate_parents_object_paths(
            schema,
            operation,
            object_parent,
            object_name,
            reverse_objects_dict,
            reverse_doc_type_dict,
            already_seen,
        )

        paths = paths.union(parent_paths)

    if settings.MAX_PATHS_PER_OBJECT > 0:
        paths = set(random.sample(paths, settings.MAX_PATHS_PER_OBJECT))

    schema[object_type][object_name]['paths'][operation] = paths  # type: ignore[index]

    return schema


def generate_objects_paths_for_operation(
    schema: IntrospectionSchema,
    operation: Operation,
) -> IntrospectionSchema:
    """Save the paths to one object in the specification."""

    if schema[operation]:
        reverse_objects_dict = _generate_reverse_objects_dict(schema)
        reverse_doc_type_dict = _generate_reverse_doc_type_dict(schema, operation)

        for object_name in schema['objects']:
            schema = _generate_nested_object_paths(
                schema,
                operation,
                object_name,
                'objects',
                reverse_objects_dict,
                reverse_doc_type_dict,
            )

        for union_name in schema['unions']:
            schema = _generate_nested_object_paths(
                schema,
                operation,
                union_name,
                'unions',
                reverse_objects_dict,
                reverse_doc_type_dict,
            )

    return schema


def generate_objects_paths(schema: IntrospectionSchema) -> IntrospectionSchema:
    """Iterate over all the operations and save the paths to one object in the specification."""

    for operation in SupportedOperations:
        schema = generate_objects_paths_for_operation(schema, operation)

    return schema



"""Internal GraphEnum entities."""
from enum import Enum


class Callbacks(Enum):

    """The callbacks that can be used."""

    PRE_SERIALIZE = 'pre_serialize'
    POST_SERIALIZE = 'post_serialize'
    PRE_SCHEMA_GENERATION = 'pre_schema_generation'
    POST_SCHEMA_GENERATION = 'post_schema_generation'



"""CLI entrypoint of the module."""

import argparse
import os
import sys
from typing import List, Optional

from graphenum.main import introspect
from graphenum.utils import dump_schema


def parse_args(args: List[str]) -> argparse.Namespace:
    """Parse the arguments."""

    parser = argparse.ArgumentParser()

    parser.add_argument(
        '-u',
        '--url',
        dest='url',
        type=str,
        help='The URL of the GraphQL endpoint.',
        default=None,
    )
    parser.add_argument(
        '-v',
        '--verbose',
        dest='verbose_mode',
        type=bool,
        help='Verbose',
        default=bool(os.getenv('DEBUG')),
    )
    parser.add_argument(
        '-s',
        '--schema-path',
        dest='schema_path',
        type=str,
        help='Output schema path',
        default=None,
    )
    parser.add_argument(
        '-o',
        '--output-schema',
        dest='output_schema',
        type=str,
        help='GraphQL schema path',
        default=None,
    )
    return parser.parse_args(args)


def cli(argv: Optional[List[str]] = None) -> None:
    """CLI entrypoint."""

    args = parse_args(argv or sys.argv[1:])
    assert args.url, 'Url is required'
    assert args.output_schema or args.verbose_mode or os.getenv('DEBUG'), 'Either schema_path or verbose_mode is required to output the schema'

    result = introspect(
        url=args.url,
        verbose_mode=args.verbose_mode,
        schema_path=args.schema_path,
    )

    schema = result[0]
    if args.output_schema:
        dump_schema(args.output_schema, schema)



"""
Each sub-section below defines the expected fields of __Type given each possible value of the __TypeKind enum:
source: https://spec.graphql.org/draft/#sec-The-__Type-Type

    x "SCALAR"
    ✓ "OBJECT"
    ✓ "INTERFACE"
    ✓ "UNION"
    ✓ "ENUM"
    ✓ "INPUT_OBJECT"
    x "LIST"
    x "NON_NULL"
"""

import sys
from typing import Dict, List, Optional, Set, Tuple

from graphenum.entities.graphql.properties import ArgName, EnumName, FieldName, GraphKind, GraphObjectType, GraphPath, GraphTypeName, GraphValue
from graphenum.entities.graphql.properties import InputObjectName, InterfaceName, ObjectName, Operation, UnionName

if sys.version_info >= (3, 8):
    from typing import TypedDict  # pylint: disable=no-name-in-module
else:
    from typing_extensions import TypedDict


class MinimalGraphType(TypedDict):

    """Minimal GraphQL Type."""

    name: GraphTypeName
    nullable: List[bool]
    depth: int  # depth = 1 <=> it's a list | depth >= 2 <=> it is a list of lists
    kind: GraphKind


class GraphType(MinimalGraphType, total=False):

    """GraphQL Type."""


class MinimalGraphField(TypedDict):

    """Minimal GraphQL Field."""

    name: FieldName
    description: str
    args: Dict[ArgName, 'GraphArg']
    type: GraphType
    isDeprecated: bool
    deprecationReason: str


class GraphField(MinimalGraphField, total=False):

    """GraphQL Field."""


class GraphObject(TypedDict):

    """GraphQL Object.

    Source:
        https://spec.graphql.org/draft/#sec-The-__Type-Type.Object
    """

    kind: GraphKind
    name: ObjectName
    description: Optional[str]
    fields: Dict[FieldName, GraphField]
    interfaces: Dict[InterfaceName, 'GraphInterface']  # type: ignore[misc]
    enumValues: Dict[str, 'GraphEnum']
    paths: Dict[Operation, Set[GraphPath]]  # GraphPath that end on the object


class GraphInterface(TypedDict):

    """GraphQL Interface.

    Source:
        https://spec.graphql.org/draft/#sec-The-__Type-Type.Interface
    """

    kind: GraphKind
    name: InterfaceName
    description: Optional[str]
    fields: Dict[FieldName, GraphField]
    interfaces: Dict[InterfaceName, 'GraphInterface']  # type: ignore[misc]
    possibleTypes: Dict[GraphObjectType, GraphType]
    paths: Dict[Operation, Set[GraphPath]]  # GraphPath that end on the object


class GraphUnion(TypedDict):

    """GraphQL Union.

    Source:
        https://spec.graphql.org/draft/#sec-The-__Type-Type.Union
    """

    kind: GraphKind
    name: UnionName
    description: Optional[str]
    possibleTypes: Dict[GraphObjectType, GraphType]
    paths: Dict[Operation, Set[GraphPath]]  # GraphPath that end on the object


class GraphArg(TypedDict):

    """Minimal GraphQL Argument."""

    name: ArgName
    description: str
    defaultValue: GraphValue
    type: GraphType
    sources: List[Tuple[ObjectName, FieldName]]


class EnumValue(TypedDict):

    """Type for enum values."""

    name: str
    description: Optional[str]
    isDeprecated: bool
    deprecationReason: Optional[str]


class GraphEnum(TypedDict):

    """GraphQL Input.

    Source:
        https://spec.graphql.org/draft/#sec-The-__Type-Type.Enum
    """

    kind: GraphKind
    name: EnumName
    description: Optional[str]
    enumValues: Dict[str, EnumValue]


class GraphInputObject(TypedDict):

    """GraphQL Input Object.

    Source:
        https://spec.graphql.org/draft/#sec-The-__Type-Type.Input-Object
    """

    kind: GraphKind
    name: InputObjectName
    description: Optional[str]
    inputFields: Dict[ArgName, GraphArg]



"""Manage sources of data for the module."""

import json
from typing import Dict, Optional, Tuple

import aiohttp

INTROSPECTION_QUERY: Dict[str, str] = {
    'query':
        """
query IntrospectionQuery {
  __schema {
    queryType {
      name
    }
    mutationType {
      name
    }
    subscriptionType {
      name
    }
    types {
      ...FullType
    }
    directives {
      name
      description
      locations
      args {
        ...InputValue
      }
    }
  }
}

fragment FullType on __Type {
  kind
  name
  description
  fields(includeDeprecated: true) {
    name
    description
    args {
      ...InputValue
    }
    type {
      ...TypeRef
    }
    isDeprecated
    deprecationReason
  }
  inputFields {
    ...InputValue
  }
  interfaces {
    ...TypeRef
  }
  enumValues(includeDeprecated: true) {
    ...EnumValue
  }
  possibleTypes {
    ...TypeRef
  }
}

fragment EnumValue on __EnumValue {
    name
    description
    isDeprecated
    deprecationReason
}

fragment InputValue on __InputValue {
  name
  description
  type {
    ...TypeRef
  }
  defaultValue
}

fragment TypeRef on __Type {
  kind
  name
  ofType {
    kind
    name
    ofType {
      kind
      name
      ofType {
        kind
        name
        ofType {
          kind
          name
          ofType {
            kind
            name
            ofType {
              kind
              name
              ofType {
                kind
                name
              }
            }
          }
        }
      }
    }
  }
}
"""
}


async def fetch_schema(
    url: str,
    headers: Optional[Dict[str, str]],
) -> Tuple[Optional[Dict], aiohttp.ClientResponse]:
    """Fetch a schema from a URL."""

    async with aiohttp.ClientSession() as session:
        async with session.post(url, json=INTROSPECTION_QUERY, headers=headers) as resp:
            try:
                body = await resp.json()
            except aiohttp.ClientError:
                return {}, resp

    data = body.get('data', body)
    assert '__schema' in data, 'Invalid schema'
    return data['__schema'], resp


def load_schema(path: str) -> Dict:
    """Load a schema from a file."""

    with open(path, 'r', encoding='utf-8') as f:
        return json.load(f)



"""Handlers for logging."""

import logging
import os


def setup_logger(verbose_mode: bool = False) -> logging.Logger:
    """Setup the logger."""

    logger = logging.getLogger('graphintrospect')
    logger.setLevel(logging.DEBUG if os.getenv('DEBUG') or verbose_mode else logging.INFO)

    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter(os.getenv('LOG_FORMAT') or '%(asctime)s %(levelname)s %(name)s - %(message)s'))

    for h in logger.handlers:
        logger.removeHandler(h)

    logger.addHandler(handler)
    logger.propagate = False

    return logger


LOG: logging.Logger = setup_logger()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['graphenum', 'graphenum.entities', 'graphenum.entities.graphql']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0']

entry_points = \
{'console_scripts': ['graphenum = graphenum:cli']}

setup_kwargs = {
    'name': 'graphenum',
    'version': '0.0.1',
    'description': 'GraphQL introspection and path enumeration.',
    'long_description': '# graphenum\n\nIntrospect a GraphQL endpoint and generate his specification schema with path enumeration.\n\n## Getting Started\n\nIt takes only two simple steps to fingerprint an endpoint using GraphDNA.\n\n```bash\npip install graphenum\ngraphenum -u https://example.com/graphql -o schema.json\n```\n\n## Options\n\nGraphEnum supports the following options:\n\n```bash\ngraphenum \\\n# url of the GraphQL endpoint\n-u/--url https://example.com/graphql \\\n# input schema file (raw GraphQL schema)\n-s/--schema-path schema.json \\\n# output file name\n-o/--output-schema schema.json \\\n# verbose output\n-v/--verbose\n```\n\n## Environment Variables\n\n**Logger** - *No effect if you pass your own logger*\n| Name | Values  | Default| Behavior|\n|------|--------|--------|--------|\n| `LOG_FORMAT` | `console`, `json` | `console` | Change the log format accordingly |\n| `DEBUG` | `True`, `False` | `False` | Enable debug logging |\n\n## Integration\n\n```python\nimport logging\n\nfrom typing import Callable, Dict, Optional, List, Tuple\nfrom graphenum import IntrospectionSchema\nfrom graphenum.entities.internal import Callbacks\n\nasync def async_introspect(\n    url: str,\n    logger: Optional[logging.Logger] = None,\n    headers: Optional[Dict[str, str]] = None,\n    verbose_mode: bool = False,\n    schema_path: Optional[str] = None,\n    callbacks: Optional[Dict[Callbacks, List[Callable]]] = None,\n) -> Tuple[IntrospectionSchema, aiohttp.ClientResponse]:\n    ...\n\ndef introspect(\n    url: str,\n    logger: Optional[logging.Logger] = None,\n    headers: Optional[Dict[str, str]] = None,\n    verbose_mode: bool = False,\n    schema_path: Optional[str] = None,\n    callbacks: Optional[Dict[Callbacks, List[Callable]]] = None,\n) -> Tuple[IntrospectionSchema, aiohttp.ClientResponse]:\n    ...\n```\n\n## Local installation\n\n```bash\ngit clone git@github.com:Escape-Technologies/graphenum.git\ncd graphenum\nchmod +x ./install-dev.sh\n./install-dev.sh\n```\n\n## Contributing\n\nPull requests are welcome. For major changes, please open an issue first to discuss what you would like to change.\n\nPlease make sure to update tests as appropriate.\n\n## License ![PyPI - License](https://img.shields.io/pypi/l/GraphDNA)\n\n[MIT](https://choosealicense.com/licenses/mit/)',
    'author': 'Escape Technologies SAS',
    'author_email': 'ping@escape.tech',
    'maintainer': 'Swan',
    'maintainer_email': 'swan@escape.tech',
    'url': 'https://escape.tech/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<=3.11',
}


setup(**setup_kwargs)
